@@ -1,7 +1,9 @@
 import numpy.lib.stride_tricks
 import numpy as np
-from numpy.fft import rfftn, irfftn
+from scipy.fft import rfftn, irfftn
 from numpy import ma
+from tqdm import tqdm
+from openpiv.pyprocess import get_field_shape, find_first_peak
 
 """This module contains a pure python implementation of the basic
 cross-correlation algorithm for PIV image processing."""
@@ -24,100 +26,89 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 """
 
 
-
-
-
-def get_coordinates(image_size, window_size, overlap):
+def get_coordinates(image_size, search_area_size, window_size, overlap):
     """Compute the x, y coordinates of the centers of the interrogation windows.
 
     Parameters
     ----------
     image_size: two elements tuple
-        a two dimensional tuple for the pixel size of the image
-        first element is number of rows, second element is
-        the number of columns.
+        a three dimensional tuple for the pixel size of the image
 
-    window_size: int
-        the size of the interrogation windows.
+    window_size: tuple
+        the size of the interrogation window.
 
-    overlap: int
+    search_area_size: tuple
+        the size of the search area window.
+
+    overlap: tuple
         the number of pixel by which two adjacent interrogation
         windows overlap.
 
 
     Returns
     -------
-    x : 2d np.ndarray
-        a two dimensional array containing the x coordinates of the
+    x : 23 np.ndarray
+        a three dimensional array containing the x coordinates of the
         interrogation window centers, in pixels.
 
-    y : 2d np.ndarray
-        a two dimensional array containing the y coordinates of the
+    y : 23 np.ndarray
+        a three dimensional array containing the y coordinates of the
+        interrogation window centers, in pixels.
+
+    z : 23 np.ndarray
+        a three dimensional array containing the y coordinates of the
         interrogation window centers, in pixels.
 
     """
 
     # get shape of the resulting flow field
-    field_shape = get_field_shape(image_size, window_size, overlap)
+    field_shape = get_field_shape(image_size, search_area_size, overlap)
 
-    # compute grid coordinates of the interrogation window centers
-    # compute grid coordinates of the interrogation window centers
-    x = np.arange(field_shape[1]) * (window_size[1] - overlap[1]) + window_size[1] / 2.0
-    y = np.arange(field_shape[0]) * (window_size[0] - overlap[0]) + window_size[0] / 2.0
+    # compute grid coordinates of the search area centers
+    x = (
+        np.arange(field_shape[1]) * (window_size[1] - overlap[1])
+        + (search_area_size[1] - 1) / 2.0
+    )
+    y = (
+        np.arange(field_shape[0]) * (window_size[0] - overlap[0])
+        + (search_area_size[0] - 1) / 2.0
+    )
+    z = (
+        np.arange(field_shape[2]) * (window_size[2] - overlap[2])
+        + (search_area_size[2] - 1) / 2.0
+    )
+
+    # moving coordinates further to the center, so that the points at the extreme left/right or top/bottom
+    # have the same distance to the window edges. For simplicity only integer movements are allowed.
+    x += (
+        image_size[1]
+        - 1
+        - (
+            (field_shape[1] - 1) * (window_size[1] - overlap[1])
+            + (search_area_size[1] - 1)
+        )
+    ) // 2
+    y += (
+        image_size[0]
+        - 1
+        - (
+            (field_shape[0] - 1) * (window_size[0] - overlap[0])
+            + (search_area_size[0] - 1)
+        )
+    ) // 2
+    z += (
+        image_size[2]
+        - 1
+        - (
+            (field_shape[2] - 1) * (window_size[2] - overlap[2])
+            + (search_area_size[2] - 1)
+        )
+    ) // 2
 
-    z = np.arange(field_shape[2]) * (window_size[2] - overlap[2]) + window_size[2] / 2.0
     return np.meshgrid(x, y, z)
 
 
-def get_field_shape(image_size, window_size, overlap):
-    """Compute the shape of the resulting flow field.
-
-    Given the image size, the interrogation window size and
-    the overlap size, it is possible to calculate the number
-    of rows and columns of the resulting flow field.
-
-    Parameters
-    ----------
-    image_size: two elements tuple
-        a two dimensional tuple for the pixel size of the image
-        first element is number of rows, second element is
-        the number of columns.
-
-    window_size: int
-        the size of the interrogation window.
-
-    overlap: int
-        the number of pixel by which two adjacent interrogation
-        windows overlap.
-
-
-    Returns
-    -------
-    field_shape : two elements tuple
-        the shape of the resulting flow field
-    """
-
-    return (np.array(image_size) - np.array(window_size)) // (np.array(window_size) - np.array(overlap)) + 1
-
-
-def find_first_peak(corr):
-    """
-    Find row and column indices of the first correlation peak.
-
-    Parameters
-    ----------
-    corr : np.ndarray
-        the correlation map
-
-    Returns
-    -------
-
-    """
-
-    return np.unravel_index(np.argmax(corr), corr.shape), corr.max()
-
-
-def find_second_peak(corr, i=None, j=None, z=None, width=2):
+def find_second_peak_3D(corr, i=None, j=None, z=None, width=2):
     """
     Find the value of the second largest peak.
 
@@ -130,8 +121,8 @@ def find_second_peak(corr, i=None, j=None, z=None, width=2):
     corr: np.ndarray
           the correlation map.
 
-    i,j : ints
-          row and column location of the first peak.
+    i,j,z : ints
+          row, column and layer location of the first peak.
 
     width : int
         the half size of the region around the first correlation
@@ -176,7 +167,7 @@ def find_second_peak(corr, i=None, j=None, z=None, width=2):
     return (i, j, z), corr_max2
 
 
-def find_subpixel_peak_position(corr, subpixel_method='gaussian'):
+def find_subpixel_peak_position(corr, subpixel_method="gaussian"):
     """
     Find subpixel approximation of the correlation peak.
 
@@ -221,28 +212,49 @@ def find_subpixel_peak_position(corr, subpixel_method='gaussian'):
         cb = corr[peak1_i, peak1_j, peak1_z + 1]
 
         # gaussian fit
-        if np.any(np.array([c, cl, cr, cd, cu, cf, cb]) < 0) and subpixel_method == 'gaussian':
-            subpixel_method = 'centroid'
+        if (
+            np.any(np.array([c, cl, cr, cd, cu, cf, cb]) < 0)
+            and subpixel_method == "gaussian"
+        ):
+            subpixel_method = "centroid"
 
         try:
-            if subpixel_method == 'centroid':
+            if subpixel_method == "centroid":
                 subp_peak_position = (
-                    ((peak1_i - 1) * cl + peak1_i * c + (peak1_i + 1) * cr) / (cl + c + cr),
-                    ((peak1_j - 1) * cd + peak1_j * c + (peak1_j + 1) * cu) / (cd + c + cu),
-                    ((peak1_z - 1) * cf + peak1_z * c + (peak1_z + 1) * cb) / (cf + c + cb))
+                    ((peak1_i - 1) * cl + peak1_i * c + (peak1_i + 1) * cr)
+                    / (cl + c + cr),
+                    ((peak1_j - 1) * cd + peak1_j * c + (peak1_j + 1) * cu)
+                    / (cd + c + cu),
+                    ((peak1_z - 1) * cf + peak1_z * c + (peak1_z + 1) * cb)
+                    / (cf + c + cb),
+                )
 
-            elif subpixel_method == 'gaussian':
-                with numpy.errstate(divide='ignore'):
+            elif subpixel_method == "gaussian":
+                with numpy.errstate(divide="ignore"):
                     subp_peak_position = (
-                        peak1_i + ((np.log(cl) - np.log(cr)) / (2 * np.log(cl) - 4 * np.log(c) + 2 * np.log(cr))),
-                        peak1_j + ((np.log(cd) - np.log(cu)) / (2 * np.log(cd) - 4 * np.log(c) + 2 * np.log(cu))),
-                        peak1_z + ((np.log(cf) - np.log(cb)) / (2 * np.log(cf) - 4 * np.log(c) + 2 * np.log(cb)))
+                        peak1_i
+                        + (
+                            (np.log(cl) - np.log(cr))
+                            / (2 * np.log(cl) - 4 * np.log(c) + 2 * np.log(cr))
+                        ),
+                        peak1_j
+                        + (
+                            (np.log(cd) - np.log(cu))
+                            / (2 * np.log(cd) - 4 * np.log(c) + 2 * np.log(cu))
+                        ),
+                        peak1_z
+                        + (
+                            (np.log(cf) - np.log(cb))
+                            / (2 * np.log(cf) - 4 * np.log(c) + 2 * np.log(cb))
+                        ),
                     )
 
-            elif subpixel_method == 'parabolic':
-                subp_peak_position = (peak1_i + (cl - cr) / (2 * cl - 4 * c + 2 * cr),
-                                      peak1_j + (cd - cu) / (2 * cd - 4 * c + 2 * cu),
-                                      peak1_z + (cf - cb) / (2 * cf - 4 * c + 2 * cb))
+            elif subpixel_method == "parabolic":
+                subp_peak_position = (
+                    peak1_i + (cl - cr) / (2 * cl - 4 * c + 2 * cr),
+                    peak1_j + (cd - cu) / (2 * cd - 4 * c + 2 * cu),
+                    peak1_z + (cf - cb) / (2 * cf - 4 * c + 2 * cb),
+                )
         except:
             subp_peak_position = default_peak_position
 
@@ -252,7 +264,7 @@ def find_subpixel_peak_position(corr, subpixel_method='gaussian'):
     return np.array(subp_peak_position) - np.array(default_peak_position)
 
 
-def sig2noise_ratio(corr, sig2noise_method='peak2peak', width=2):
+def sig2noise_ratio(corr, sig2noise_method="peak2peak", width=2):
     """
     Computes the signal to noise ratio from the correlation map.
 
@@ -286,23 +298,29 @@ def sig2noise_ratio(corr, sig2noise_method='peak2peak', width=2):
     (peak1_i, peak1_j, peak1_z), corr_max1 = find_first_peak(corr)
 
     # now compute signal to noise ratio
-    if sig2noise_method == 'peak2peak':
+    if sig2noise_method == "peak2peak":
         # find second peak height
-        (peak1_i, peak1_j, peak1_z), corr_max2 = find_second_peak(
-            corr, peak1_i, peak1_j, peak1_z, width=width)
+        (peak1_i, peak1_j, peak1_z), corr_max2 = find_second_peak_3D(
+            corr, peak1_i, peak1_j, peak1_z, width=width
+        )
 
         # if it's an empty interrogation window
         # if the image is lacking particles, totally black it will correlate to very low value, but not zero
         # if the first peak is on the borders, the correlation map is also wrong
-        if corr_max1 < 1e-3 or any([x == 0 or x == corr.shape[i] for i, x in enumerate([peak1_i, peak1_j, peak1_z])]):
+        if corr_max1 < 1e-3 or any(
+            [
+                x == 0 or x == corr.shape[i]
+                for i, x in enumerate([peak1_i, peak1_j, peak1_z])
+            ]
+        ):
             return 0.0
 
-    elif sig2noise_method == 'peak2mean':
+    elif sig2noise_method == "peak2mean":
         # find mean of the correlation map
         corr_max2 = corr.mean()
 
     else:
-        raise ValueError('wrong sig2noise_method')
+        raise ValueError("wrong sig2noise_method")
 
     # avoid dividing by zero
     try:
@@ -313,7 +331,9 @@ def sig2noise_ratio(corr, sig2noise_method='peak2peak', width=2):
     return sig2noise
 
 
-def correlate_windows(window_a, window_b, corr_method='fft', nfftx=None, nffty=None, nfftz=None):
+def correlate_windows(
+    window_a, window_b, correlation_method="fft", nfftx=None, nffty=None, nfftz=None
+):
     """Compute correlation function between two interrogation windows.
 
     The correlation function can be computed by using the correlation
@@ -327,7 +347,7 @@ def correlate_windows(window_a, window_b, corr_method='fft', nfftx=None, nffty=N
     window_b : 2d np.ndarray
         a two dimensions array for the second interrogation window.
 
-    corr_method   : string
+    correlation_method   : string
         one method is currently implemented: 'fft'.
 
     nfftx   : int
@@ -357,7 +377,7 @@ def correlate_windows(window_a, window_b, corr_method='fft', nfftx=None, nffty=N
     It leads to inconsistency of the output
     """
 
-    if corr_method == 'fft':
+    if correlation_method == "fft":
         window_b = np.conj(window_b[::-1, ::-1, ::-1])
         if nfftx is None:
             nfftx = nextpower2(window_b.shape[0] + window_a.shape[0])
@@ -366,20 +386,20 @@ def correlate_windows(window_a, window_b, corr_method='fft', nfftx=None, nffty=N
         if nfftz is None:
             nfftz = nextpower2(window_b.shape[2] + window_a.shape[2])
 
-        f2a = rfftn(normalize_intensity(window_a),
-                    s=(nfftx, nffty, nfftz))
-        f2b = rfftn(normalize_intensity(window_b),
-                    s=(nfftx, nffty, nfftz))
+        f2a = rfftn(normalize_intensity(window_a), s=(nfftx, nffty, nfftz))
+        f2b = rfftn(normalize_intensity(window_b), s=(nfftx, nffty, nfftz))
         corr = irfftn(f2a * f2b).real
-        corr = corr[:window_a.shape[0] + window_b.shape[0],
-                    :window_b.shape[1] + window_a.shape[1],
-                    :window_b.shape[2] + window_a.shape[2]]
+        corr = corr[
+            : window_a.shape[0] + window_b.shape[0],
+            : window_b.shape[1] + window_a.shape[1],
+            : window_b.shape[2] + window_a.shape[2],
+        ]
         return corr
-    # elif corr_method == 'direct':
+    # elif correlation_method == 'direct':
     #     return convolve2d(normalize_intensity(window_a),
     #                       normalize_intensity(window_b[::-1, ::-1, ::-1]), 'full')
     else:
-        raise ValueError('method is not implemented')
+        raise ValueError("method is not implemented")
 
 
 def normalize_intensity(window):
@@ -409,34 +429,40 @@ def nextpower2(i):
 
 def check_input(window_size, overlap, search_area_size, frame_a, frame_b):
     # check the inputs for validity
-    search_area_size = [ws if x == 0 or x is None else x for x, ws in zip(search_area_size, window_size)]
+    search_area_size = [
+        ws if x == 0 or x is None else x for x, ws in zip(search_area_size, window_size)
+    ]
 
-    if any((np.array(window_size) - np.array(overlap)) < 0):
-        raise ValueError('Overlap has to be smaller than the window_size')
+    if any((np.array(window_size) - np.array(overlap)) <= 0):
+        raise ValueError("Overlap has to be smaller than the window_size")
 
     if any((np.array(search_area_size) - np.array(window_size)) < 0):
-        raise ValueError('Search size cannot be smaller than the window_size')
+        raise ValueError("Search size cannot be smaller than the window_size")
 
     if any([ws > ims for ws, ims in zip(window_size, frame_a.shape)]):
-        raise ValueError('window size cannot be larger than the image')
+        raise ValueError("window size cannot be larger than the image")
 
     if any([ims_a != ims_b for ims_a, ims_b in zip(frame_a.shape, frame_b.shape)]):
-        raise ValueError('frame a and frame b have different sizes.')
+        raise ValueError("frame a and frame b have different sizes.")
 
     return window_size, overlap, search_area_size
 
 
 def extended_search_area_piv3D(
-        frame_a, frame_b,
-        window_size,
-        overlap=0,
-        dt=(1.0, 1.0, 1.0),
-        search_area_size=None,
-        corr_method='fft',
-        subpixel_method='gaussian',
-        sig2noise_method=None,
-        width=2,
-        nfftx=None, nffty=None, nfftz=None):
+    frame_a,
+    frame_b,
+    window_size,
+    overlap=(0, 0, 0),
+    dt=(1.0, 1.0, 1.0),
+    search_area_size=None,
+    correlation_method="fft",
+    subpixel_method="gaussian",
+    sig2noise_method=None,
+    width=2,
+    nfftx=None,
+    nffty=None,
+    nfftz=None,
+):
     """Standard PIV cross-correlation algorithm, with an option for
     extended area search that increased dynamic range. The search region
     in the second frame is larger than the interrogation window size in the
@@ -466,7 +492,7 @@ def extended_search_area_piv3D(
     dt : tuple
         the time delay separating the two frames [default: 1.0].
 
-    corr_method : string
+    correlation_method : string
         only one method is currently implemented: 'fft'
 
     subpixel_method : string
@@ -516,17 +542,19 @@ def extended_search_area_piv3D(
         a three dimensional array containing the w velocity component,
         in pixels/seconds.
 
-    sig2noise : 2d np.ndarray, ( optional: only if sig2noise_method is not None )
-        a two dimensional array the signal to noise ratio for each
+    sig2noise : 3d np.ndarray, (optional: only if sig2noise_method is not None)
+        a three dimensional array the signal to noise ratio for each
         window pair.
 
     """
 
     # checking if the input is correct
-    window_size, overlap, search_area_size = check_input(window_size, overlap, search_area_size, frame_a, frame_b)
+    window_size, overlap, search_area_size = check_input(
+        window_size, overlap, search_area_size, frame_a, frame_b
+    )
 
     # get field shape
-    field_shape = get_field_shape((frame_a.shape), window_size, overlap)
+    field_shape = get_field_shape(frame_a.shape, search_area_size, overlap)
 
     u = np.zeros(field_shape)
     v = np.zeros(field_shape)
@@ -536,67 +564,108 @@ def extended_search_area_piv3D(
     if sig2noise_method is not None:
         sig2noise = np.zeros(field_shape)
 
+    # shift for x and y coordinates of the search area windows so that the centers of search area windows have
+    # the same distances to the image edge at all sides. For simplicity only shifts by integers are allowed
+    x_centering = (
+        frame_a.shape[1]
+        - 1
+        - (
+            (field_shape[1] - 1) * (window_size[1] - overlap[1])
+            + (search_area_size[1] - 1)
+        )
+    ) // 2
+    y_centering = (
+        frame_a.shape[0]
+        - 1
+        - (
+            (field_shape[0] - 1) * (window_size[0] - overlap[0])
+            + (search_area_size[0] - 1)
+        )
+    ) // 2
+    z_centering = (
+        frame_a.shape[2]
+        - 1
+        - (
+            (field_shape[2] - 1) * (window_size[2] - overlap[2])
+            + (search_area_size[2] - 1)
+        )
+    ) // 2
+
     # loop over the interrogation windows
     # i, j are the row, column indices of the center of each interrogation
     # window
-    for k in range(field_shape[0]):
-        # range(range(search_area_size/2, frame_a.shape[0] - search_area_size/2, window_size - overlap ):
+    for k in tqdm(range(field_shape[0])):
         for m in range(field_shape[1]):
-            # range(search_area_size/2, frame_a.shape[1] - search_area_size/2 , window_size - overlap ):
             for l in range(field_shape[2]):
-                # Select first the largest window, work like usual from the top left corner
-                # the left edge goes as:
-                # e.g. 0, (search_area_size - overlap), 2*(search_area_size - overlap),....
 
-                il = k * (search_area_size[0] - overlap[0])
-                ir = il + search_area_size[0]
+                # centers of search area. (window_size - overlap) defines the distance between each center
+                # and (search_area_size - 1)/2.0 moves the center points away from the left or top image edge
+                y = k * (window_size[0] - overlap[0]) + (search_area_size[0] - 1) / 2.0
+                x = m * (window_size[1] - overlap[1]) + (search_area_size[1] - 1) / 2.0
+                z = l * (window_size[2] - overlap[2]) + (search_area_size[2] - 1) / 2.0
 
-                # same for top-bottom
-                jt = m * (search_area_size[1] - overlap[1])
-                jb = jt + search_area_size[1]
+                # moving the coordinates a bit to the center, to guarantee that the distance of a extreme
+                # point at the image edges is symmetric all all edges
+                x += x_centering
+                y += y_centering
+                z += z_centering
 
-                # same for front-back
-                jf = l * (search_area_size[2] - overlap[2])
-                jba = jf + search_area_size[2]
+                # left, right, top, bottom, front, back indices of the search area edges
+                # note that x - (search_area_size +/- 1)/2  always returns an integer due to the definition of x and y
+                # see also "get_coordinates()"
+                il = int(y - (search_area_size[0] - 1) / 2.0)
+                ir = int(y + (search_area_size[0] + 1) / 2.0)
+                it = int(x - (search_area_size[1] - 1) / 2.0)
+                ib = int(x + (search_area_size[1] + 1) / 2.0)
+                ifr = int(z - (search_area_size[2] - 1) / 2.0)
+                iba = int(z + (search_area_size[2] + 1) / 2.0)
+                # picking the search area from frame b
+                window_b = frame_b[il:ir, it:ib, ifr:iba]
 
-                # pick up the window in the second image
-                window_b = frame_b[il:ir, jt:jb, jf:jba]
+                # left, right, top, bottom, front, back indices of the interrogation window
+                # Sometimes the interrogation window cannot be placed in the middle of the search area, e.g.
+                # in the case of window_size=3 search_area_size=4. In this case the interrogation window
+                # is shifted 0.5 pixels to the left/top, which is achieved by rounding the indices
+                #  down during the int() conversion
+                il = int(y - (window_size[0] - 1) / 2)
+                ir = int(y + (window_size[0] + 1) / 2)
+                it = int(x - (window_size[1] - 1) / 2)
+                ib = int(x + (window_size[1] + 1) / 2)
+                ifr = int(z - (window_size[2] - 1) / 2)
+                iba = int(z + (window_size[2] + 1) / 2)
+                # picking the interrogation window from frame a
+                window_a = frame_a[il:ir, it:ib, ifr:iba]
 
-                # now shift the left corner of the smaller window inside the larger one
-                il += (search_area_size[0] - window_size[0]) // 2
-                # and it's right side is just a window_size apart
-                ir = il + window_size[0]
-                # same same
-                jt += (search_area_size[1] - window_size[1]) // 2
-                jb = jt + window_size[1]
-                jf += (search_area_size[2] - window_size[2]) // 2
-                jba = jf + window_size[2]
-
-                window_a = frame_a[il:ir, jt:jb, jf:jba]
-
-                if np.any(window_a) and np.any(window_b):
-                    corr = correlate_windows(window_a, window_b,
-                                             corr_method=corr_method,
-                                             nfftx=nfftx, nffty=nffty, nfftz=nfftz)
+                if np.any(window_a):
+                    corr = correlate_windows(
+                        window_a,
+                        window_b,
+                        correlation_method=correlation_method,
+                        nfftx=nfftx,
+                        nffty=nffty,
+                        nfftz=nfftz,
+                    )
 
                     # get subpixel approximation for peak position row and column index
-                    row, col, z = find_subpixel_peak_position(corr, subpixel_method=subpixel_method)
+                    row, col, z = find_subpixel_peak_position(
+                        corr, subpixel_method=subpixel_method
+                    )
 
                     row -= (search_area_size[0] + window_size[0] - 1) // 2
                     col -= (search_area_size[1] + window_size[1] - 1) // 2
                     z -= (search_area_size[2] + window_size[2] - 1) // 2
 
                     # get displacements, apply coordinate system definition
-                    u[k, m, l], v[k, m, l], w[k, m, l] = -col, row, z
+                    u[k, m, l], v[k, m, l], w[k, m, l] = col, -row, -z
 
                     # get signal to noise ratio
                     if sig2noise_method is not None:
                         sig2noise[k, m, l] = sig2noise_ratio(
-                            corr, sig2noise_method=sig2noise_method, width=width)
+                            corr, sig2noise_method=sig2noise_method, width=width
+                        )
 
     # return output if user wanted sig2noise information
     if sig2noise_method is not None:
         return u / dt[0], v / dt[1], w / dt[2], sig2noise
     else:
         return u / dt[0], v / dt[1], w / dt[2]
-
